@@ -35,7 +35,6 @@
 #include "buffer-lcm.h"
 #include "copy.h"
 #include "cp-hash.h"
-#include "dirname.h"
 #include "euidaccess.h"
 #include "error.h"
 #include "fcntl--.h"
@@ -261,7 +260,7 @@ copy_reg (char const *src_name, char const *dst_name,
       return false;
     }
 
-  if (fstat (source_desc, &src_open_sb))
+  if (fstat (source_desc, &src_open_sb) != 0)
     {
       error (0, errno, _("cannot fstat %s"), quote (src_name));
       return_val = false;
@@ -281,11 +280,7 @@ copy_reg (char const *src_name, char const *dst_name,
 
   /* These semantics are required for cp.
      The if-block will be taken in move_mode.  */
-  if (*new_dst)
-    {
-      dest_desc = open (dst_name, O_WRONLY | O_CREAT | O_BINARY, dst_mode);
-    }
-  else
+  if (! *new_dst)
     {
       dest_desc = open (dst_name, O_WRONLY | O_TRUNC | O_BINARY, dst_mode);
 
@@ -297,15 +292,17 @@ copy_reg (char const *src_name, char const *dst_name,
 	      return_val = false;
 	      goto close_src_desc;
 	    }
+	  if (x->verbose)
+	    printf (_("removed %s\n"), quote (dst_name));
 
 	  /* Tell caller that the destination file was unlinked.  */
 	  *new_dst = true;
-
-	  /* Try the open again, but this time with different flags.  */
-	  dest_desc = open (dst_name, O_WRONLY | O_CREAT | O_BINARY, dst_mode);
 	}
     }
 
+  if (*new_dst)
+    dest_desc = open (dst_name, O_WRONLY | O_CREAT | O_BINARY, dst_mode);
+
   if (dest_desc < 0)
     {
       error (0, errno, _("cannot create regular file %s"), quote (dst_name));
@@ -313,7 +310,7 @@ copy_reg (char const *src_name, char const *dst_name,
       goto close_src_desc;
     }
 
-  if (fstat (dest_desc, &sb))
+  if (fstat (dest_desc, &sb) != 0)
     {
       error (0, errno, _("cannot fstat %s"), quote (dst_name));
       return_val = false;
@@ -933,6 +930,18 @@ abandon_move (const struct cp_options *x,
               && ! yesno ()));
 }
 
+/* Print --verbose output on standard output, e.g. `new' -> `old'.
+   If BACKUP_DST_NAME is non-NULL, then also indicate that it is
+   the name of a backup file.  */
+static void
+emit_verbose (char const *src, char const *dst, char const *backup_dst_name)
+{
+  printf ("%s -> %s", quote_n (0, src), quote_n (1, dst));
+  if (backup_dst_name)
+    printf (_(" (backup: %s)"), quote (backup_dst_name));
+  putchar ('\n');
+}
+
 /* Copy the file SRC_NAME to the file DST_NAME.  The files may be of
    any type.  NEW_DST should be true if the file DST_NAME cannot
    exist because its parent directory was just created; NEW_DST should
@@ -944,7 +953,6 @@ abandon_move (const struct cp_options *x,
    Set *COPY_INTO_SELF if SRC_NAME is a parent of (or the
    same as) DST_NAME; otherwise, clear it.
    Return true if successful.  */
-
 static bool
 copy_internal (char const *src_name, char const *dst_name,
 	       bool new_dst,
@@ -958,7 +966,6 @@ copy_internal (char const *src_name, char const *dst_name,
   struct stat src_sb;
   struct stat dst_sb;
   mode_t src_mode;
-  mode_t src_type;
   mode_t dst_mode IF_LINT (= 0);
   bool restore_dst_mode = false;
   char *earlier_file = NULL;
@@ -979,11 +986,9 @@ copy_internal (char const *src_name, char const *dst_name,
       return false;
     }
 
-  src_type = src_sb.st_mode;
-
   src_mode = src_sb.st_mode;
 
-  if (S_ISDIR (src_type) && !x->recursive)
+  if (S_ISDIR (src_mode) && !x->recursive)
     {
       error (0, 0, _("omitting directory %s"), quote (src_name));
       return false;
@@ -1022,40 +1027,66 @@ copy_internal (char const *src_name, char const *dst_name,
 	    }
 	}
       else
-	{
+	{ /* Here, we know that dst_name exists, at least to the point
+	     that it is XSTAT'able.  */
 	  bool return_now;
 	  bool unlink_src;
-	  bool ok = same_file_ok (src_name, &src_sb, dst_name, &dst_sb,
-				  x, &return_now, &unlink_src);
-	  if (unlink_src)
-	    {
-	      if (!abandon_move (x, dst_name, &dst_sb)
-		  && unlink (src_name) != 0)
-		{
-		  error (0, errno, _("cannot remove %s"), quote (src_name));
-		  return false;
-		}
-	      /* Tell the caller that there's no need to remove src_name.  */
-	      if (rename_succeeded)
-		*rename_succeeded = true;
-	    }
 
-	  if (return_now)
-	    return true;
-
-	  if (! ok)
+	  if (! same_file_ok (src_name, &src_sb, dst_name, &dst_sb,
+			      x, &return_now, &unlink_src))
 	    {
 	      error (0, 0, _("%s and %s are the same file"),
 		     quote_n (0, src_name), quote_n (1, dst_name));
 	      return false;
 	    }
 
+	  /* When there is an existing destination file, we may end up
+	     returning early, and hence not copying/moving the file.
+	     This may be due to an interactive `negative' reply to the
+	     prompt about the existing file.  It may also be due to the
+	     use of the --reply=no option.
+
+	     cp and mv treat -i and -f differently.  */
+	  if (x->move_mode)
+	    {
+	      if (abandon_move (x, dst_name, &dst_sb)
+		  || (unlink_src && unlink (src_name) == 0))
+		{
+		  /* Pretend the rename succeeded, so the caller (mv)
+		     doesn't end up removing the source file.  */
+		  if (rename_succeeded)
+		    *rename_succeeded = true;
+		  if (unlink_src && x->verbose)
+		    printf (_("removed %s\n"), quote (src_name));
+		  return true;
+		}
+	      if (unlink_src)
+		{
+		  error (0, errno, _("cannot remove %s"), quote (src_name));
+		  return false;
+		}
+	    }
+	  else
+	    {
+	      if (! S_ISDIR (src_mode)
+		  && (x->interactive == I_ALWAYS_NO
+		      || (x->interactive == I_ASK_USER
+			  && (overwrite_prompt (dst_name, &dst_sb), 1)
+			  && ! yesno ())))
+		return true;
+	    }
+
+	  if (return_now)
+	    return true;
+
 	  if (!S_ISDIR (dst_sb.st_mode))
 	    {
-	      if (S_ISDIR (src_type))
+	      if (S_ISDIR (src_mode))
 		{
 		  if (x->move_mode && x->backup_type != no_backups)
 		    {
+		      /* Moving a directory onto an existing
+			 non-directory is ok only with --backup.  */
 		    }
 		  else
 		    {
@@ -1084,12 +1115,14 @@ copy_internal (char const *src_name, char const *dst_name,
 		}
 	    }
 
-	  if (!S_ISDIR (src_type))
+	  if (!S_ISDIR (src_mode))
 	    {
 	      if (S_ISDIR (dst_sb.st_mode))
 		{
 		  if (x->move_mode && x->backup_type != no_backups)
 		    {
+		      /* Moving a non-directory onto an existing
+			 directory is ok only with --backup.  */
 		    }
 		  else
 		    {
@@ -1125,37 +1158,6 @@ copy_internal (char const *src_name, char const *dst_name,
 		}
 	    }
 
-	  /* When there is an existing destination file, we may end up
-	     returning early, and hence not copying/moving the file.
-	     This may be due to an interactive `negative' reply to the
-	     prompt about the existing file.  It may also be due to the
-	     use of the --reply=no option.  */
-	  if (!S_ISDIR (src_type))
-	    {
-	      /* cp and mv treat -i and -f differently.  */
-	      if (x->move_mode)
-		{
-		  if (abandon_move (x, dst_name, &dst_sb))
-		    {
-		      /* Pretend the rename succeeded, so the caller (mv)
-			 doesn't end up removing the source file.  */
-		      if (rename_succeeded)
-			*rename_succeeded = true;
-		      return true;
-		    }
-		}
-	      else
-		{
-		  if (x->interactive == I_ALWAYS_NO
-		      || (x->interactive == I_ASK_USER
-			  && (overwrite_prompt (dst_name, &dst_sb), 1)
-			  && ! yesno ()))
-		    {
-		      return true;
-		    }
-		}
-	    }
-
 	  if (x->move_mode)
 	    {
 	      /* Don't allow user to move a directory onto a non-directory.  */
@@ -1169,9 +1171,16 @@ copy_internal (char const *src_name, char const *dst_name,
 		}
 	    }
 
-	  bool backup_directories = true;
 	  if (x->backup_type != no_backups
-	      && (!S_ISDIR (dst_sb.st_mode) || backup_directories))
+	      /* Don't try to back up a destination if the last
+		 component of src_name is "." or "..".  */
+	      && ! dot_or_dotdot (last_component (src_name))
+	      /* Create a backup of each destination directory in move mode,
+		 but not in copy mode.  FIXME: it might make sense to add an
+		 option to suppress backup creation also for move mode.
+		 That would let one use mv to merge new content into an
+		 existing hierarchy.  */
+	      && (x->move_mode || ! S_ISDIR (dst_sb.st_mode)))
 	    {
 	      char *tmp_backup = find_backup_file_name (dst_name,
 							x->backup_type);
@@ -1234,6 +1243,8 @@ copy_internal (char const *src_name, char const *dst_name,
 		  return false;
 		}
 	      new_dst = true;
+	      if (x->verbose)
+		printf (_("removed %s\n"), quote (dst_name));
 	    }
 	}
     }
@@ -1241,13 +1252,8 @@ copy_internal (char const *src_name, char const *dst_name,
   /* If the source is a directory, we don't always create the destination
      directory.  So --verbose should not announce anything until we're
      sure we'll create a directory. */
-  if (x->verbose && !S_ISDIR (src_type))
-    {
-      printf ("%s -> %s", quote_n (0, src_name), quote_n (1, dst_name));
-      if (backup_succeeded)
-	printf (_(" (backup: %s)"), quote (dst_backup));
-      putchar ('\n');
-    }
+  if (x->verbose && !S_ISDIR (src_mode))
+    emit_verbose (src_name, dst_name, backup_succeeded ? dst_backup : NULL);
 
   /* Associate the destination file name with the source device and inode
      so that if we encounter a matching dev/ino pair in the source tree
@@ -1289,7 +1295,7 @@ copy_internal (char const *src_name, char const *dst_name,
 		|| (command_line_arg
 		    && x->dereference == DEREF_COMMAND_LINE_ARGUMENTS)
 		|| x->dereference == DEREF_ALWAYS))
-	   || (x->recursive && S_ISDIR (src_type)))
+	   || (x->recursive && S_ISDIR (src_mode)))
     {
       earlier_file = remember_copied (dst_name, src_sb.st_ino, src_sb.st_dev);
     }
@@ -1302,7 +1308,7 @@ copy_internal (char const *src_name, char const *dst_name,
       /* Avoid damaging the destination file system by refusing to preserve
 	 hard-linked directories (which are found at least in Netapp snapshot
 	 directories).  */
-      if (S_ISDIR (src_type))
+      if (S_ISDIR (src_mode))
 	{
 	  /* If src_name and earlier_file refer to the same directory entry,
 	     then warn about copying a directory into itself.  */
@@ -1343,6 +1349,8 @@ copy_internal (char const *src_name, char const *dst_name,
 		  error (0, errno, _("cannot remove %s"), quote (dst_name));
 		  goto un_backup;
 		}
+	      if (x->verbose)
+		printf (_("removed %s\n"), quote (dst_name));
 	      link_failed = (link (earlier_file, dst_name) != 0);
 	    }
 
@@ -1361,8 +1369,10 @@ copy_internal (char const *src_name, char const *dst_name,
     {
       if (rename (src_name, dst_name) == 0)
 	{
-	  if (x->verbose && S_ISDIR (src_type))
-	    printf ("%s -> %s\n", quote_n (0, src_name), quote_n (1, dst_name));
+	  if (x->verbose && S_ISDIR (src_mode))
+	    emit_verbose (src_name, dst_name,
+			  backup_succeeded ? dst_backup : NULL);
+
 	  if (rename_succeeded)
 	    *rename_succeeded = true;
 
@@ -1463,7 +1473,7 @@ copy_internal (char const *src_name, char const *dst_name,
      In such cases, set this variable to zero.  */
   preserve_metadata = true;
 
-  if (S_ISDIR (src_type))
+  if (S_ISDIR (src_mode))
     {
       struct dir_list *dir;
 
@@ -1488,7 +1498,11 @@ copy_internal (char const *src_name, char const *dst_name,
 
       if (new_dst || !S_ISDIR (dst_sb.st_mode))
 	{
-	  if (mkdir (dst_name, src_mode) != 0)
+	  /* POSIX says mkdir's behavior is implementation-defined when
+	     (src_mode & ~S_IRWXUGO) != 0.  However, common practice is
+	     to ask mkdir to copy all the CHMOD_MODE_BITS, letting mkdir
+	     decide what to do with S_ISUID | S_ISGID | S_ISVTX.  */
+	  if (mkdir (dst_name, src_mode & CHMOD_MODE_BITS) != 0)
 	    {
 	      error (0, errno, _("cannot create directory %s"),
 		     quote (dst_name));
@@ -1526,7 +1540,7 @@ copy_internal (char const *src_name, char const *dst_name,
 	  remember_copied (dst_name, dst_sb.st_ino, dst_sb.st_dev);
 
 	  if (x->verbose)
-	    printf ("%s -> %s\n", quote_n (0, src_name), quote_n (1, dst_name));
+	    emit_verbose (src_name, dst_name, NULL);
 	}
 
       /* Are we crossing a file system boundary?  */
@@ -1544,7 +1558,6 @@ copy_internal (char const *src_name, char const *dst_name,
 	  delayed_ok = false;
 	}
     }
-#ifdef S_ISLNK
   else if (x->symbolic_link)
     {
       preserve_metadata = false;
@@ -1563,8 +1576,8 @@ copy_internal (char const *src_name, char const *dst_name,
 			    /* If either stat call fails, it's ok not to report
 			       the failure and say dst_name is in the current
 			       directory.  Other things will fail later.  */
-			    || stat (".", &dot_sb)
-			    || stat (dst_parent, &dst_parent_sb)
+			    || stat (".", &dot_sb) != 0
+			    || stat (dst_parent, &dst_parent_sb) != 0
 			    || SAME_INODE (dot_sb, dst_parent_sb));
 	  free (dst_parent);
 
@@ -1583,7 +1596,6 @@ copy_internal (char const *src_name, char const *dst_name,
 	  goto un_backup;
 	}
     }
-#endif
 
   else if (x->hard_link
 #ifdef LINK_FOLLOWS_SYMLINKS
@@ -1608,41 +1620,39 @@ copy_internal (char const *src_name, char const *dst_name,
 	  goto un_backup;
 	}
     }
-  else if (S_ISREG (src_type)
-	   || (x->copy_as_regular && !S_ISLNK (src_type)))
+  else if (S_ISREG (src_mode)
+	   || (x->copy_as_regular && !S_ISLNK (src_mode)))
     {
       copied_as_regular = true;
       /* POSIX says the permission bits of the source file must be
-	 used as the 3rd argument in the open call, but that's not consistent
-	 with historical practice.  */
-      if (! copy_reg (src_name, dst_name, x, src_mode, &new_dst, &src_sb))
+	 used as the 3rd argument in the open call.  Historical
+	 practice passed all the source mode bits to 'open', but the extra
+	 bits were ignored, so it should be the same either way.  */
+      if (! copy_reg (src_name, dst_name, x, src_mode & S_IRWXUGO,
+		      &new_dst, &src_sb))
 	goto un_backup;
     }
-  else
-#ifdef S_ISFIFO
-  if (S_ISFIFO (src_type))
+  else if (S_ISFIFO (src_mode))
     {
-      if (mkfifo (dst_name, src_mode))
+      /* Use mknod, rather than mkfifo, because the former preserves
+	 the special mode bits of a fifo on Solaris 10, while mkfifo
+	 does not.  */
+      if (mknod (dst_name, src_mode, 0) != 0)
 	{
 	  error (0, errno, _("cannot create fifo %s"), quote (dst_name));
 	  goto un_backup;
 	}
     }
-  else
-#endif
-    if (S_ISBLK (src_type) || S_ISCHR (src_type)
-	|| S_ISSOCK (src_type))
+  else if (S_ISBLK (src_mode) || S_ISCHR (src_mode) || S_ISSOCK (src_mode))
     {
-      if (mknod (dst_name, src_mode, src_sb.st_rdev))
+      if (mknod (dst_name, src_mode, src_sb.st_rdev) != 0)
 	{
 	  error (0, errno, _("cannot create special file %s"),
 		 quote (dst_name));
 	  goto un_backup;
 	}
     }
-  else
-#ifdef S_ISLNK
-  if (S_ISLNK (src_type))
+  else if (S_ISLNK (src_mode))
     {
       char *src_link_val = xreadlink (src_name, src_sb.st_size);
       if (src_link_val == NULL)
@@ -1686,7 +1696,7 @@ copy_internal (char const *src_name, char const *dst_name,
 	{
 	  /* Preserve the owner and group of the just-`copied'
 	     symbolic link, if possible.  */
-# if HAVE_LCHOWN
+#if HAVE_LCHOWN
 	  if (lchown (dst_name, src_sb.st_uid, src_sb.st_gid) != 0
 	      && ! chown_failure_ok (x))
 	    {
@@ -1694,16 +1704,15 @@ copy_internal (char const *src_name, char const *dst_name,
 		     dst_name);
 	      goto un_backup;
 	    }
-# else
+#else
 	  /* Can't preserve ownership of symlinks.
 	     FIXME: maybe give a warning or even error for symlinks
 	     in directories with the sticky bit set -- there, not
 	     preserving owner/group is a potential security problem.  */
-# endif
+#endif
 	}
     }
   else
-#endif
     {
       error (0, 0, _("%s has unknown file type"), quote (src_name));
       goto un_backup;
