@@ -1,5 +1,5 @@
 /* GNU's who.
-   Copyright (C) 1992-2005 Free Software Foundation, Inc.
+   Copyright (C) 1992-2006 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -48,10 +48,6 @@
 # define MAXHOSTNAMELEN 64
 #endif
 
-#ifndef S_IWGRP
-# define S_IWGRP 020
-#endif
-
 #ifdef RUN_LVL
 # define UT_TYPE_RUN_LVL(U) UT_TYPE_EQ (U, RUN_LVL)
 #else
@@ -171,7 +167,7 @@ static struct option const longopts[] = {
   {"count", no_argument, NULL, 'q'},
   {"dead", no_argument, NULL, 'd'},
   {"heading", no_argument, NULL, 'H'},
-  {"idle", no_argument, NULL, 'i'},
+  {"idle", no_argument, NULL, 'i'}, /* FIXME: deprecated: remove in late 2006 */
   {"login", no_argument, NULL, 'l'},
   {"lookup", no_argument, NULL, LOOKUP_OPTION},
   {"message", no_argument, NULL, 'T'},
@@ -324,7 +320,9 @@ print_user (const STRUCT_UTMP *utmp_ent, time_t boottime)
   char mesg;
   char idlestr[IDLESTR_LEN + 1];
   static char *hoststr;
+#if HAVE_UT_HOST
   static size_t hostlen;
+#endif
 
 #define DEV_DIR_WITH_TRAILING_SLASH "/dev/"
 #define DEV_DIR_LEN (sizeof (DEV_DIR_WITH_TRAILING_SLASH) - 1)
@@ -418,7 +416,7 @@ print_user (const STRUCT_UTMP *utmp_ent, time_t boottime)
 	  hostlen = 1;
 	  hoststr = xrealloc (hoststr, hostlen);
 	}
-      stpcpy (hoststr, "");
+      *hoststr = '\0';
     }
 #endif
 
@@ -641,8 +639,6 @@ usage (int status)
   -H, --heading     print line of column headings\n\
 "), stdout);
       fputs (_("\
-  -i, --idle        add idle time as HOURS:MINUTES, . or old\n\
-                    (deprecated, use -u)\n\
   -l, --login       print system login processes\n\
 "), stdout);
       fputs (_("\
